import numpy as np


def correlation_from_covariance(covariance):
    # see https://gist.github.com/wiso/ce2a9919ded228838703c1c7c7dad13b
    v = np.sqrt(np.diag(covariance))
    return covariance / np.outer(v, v)


def bilinear(A, x):
    return np.linalg.multi_dot((x, A, x))


def sub(A, idx):
    return A[idx, :][:, idx]


def dist(cov):
    cor = correlation_from_covariance(cov)
    # This is DANGEROUS! 1.0 - 1.0 could easily be < 0
    dist = ((1.0 - cor) / 2.)

    # problem here with 1.0 - corr...
    dist[dist < 0] = 0.0
    return dist ** .5



import numpy as np
from pyhrp.linalg import bilinear, sub


def split(v_left, v_right):
    alpha = 1 - v_left / (v_left + v_right)
    return alpha, 1 - alpha


def __hrp(node, cov, weights):

    if node.is_leaf():
        return cov[node.id][node.id], weights
    else:
        # compute the variance of the left branch
        v_left, _ = __hrp(node.left, cov, weights)

        # compute the variance of the right branch
        v_right, _ = __hrp(node.right, cov, weights)

        # compute the split factors alpha[0] and alpha[1]
        # the split is such that v_left * alpha == v_right * beta
        # shouldn't the split be v_left * alpha^2 == v_right * beta^2
        # and alpha + beta = 1
        alpha, beta = split(v_left, v_right)

        # compile a list of reachable leafs from the left node and from the right node
        # this could be done with an expensive recursive function but scipy's tree provides a powerful pre_order
        left = node.left.pre_order(lambda x: x.id)
        right = node.right.pre_order(lambda x: x.id)

        # update the weights linked to those leafs
        weights[left] = alpha * weights[left]
        weights[right] = beta * weights[right]

        # this is the list of all reachable leafs from both nodes
        idx = left + right

        # return the variance for the node and the updated weights
        return bilinear(x=weights[idx], A=sub(cov, idx=idx)), weights


def hrp_feed(node, cov):
    return __hrp(node, cov, weights=np.ones(cov.shape[1]))




import numpy as np

from pyhrp.cluster import root
from pyhrp.hrp import hrp_feed, split
import numpy.testing as nt

from pyhrp.linalg import dist


def test_dist():
    cov = np.array([[1.0, 0.2], [0.2, 2.0]])
    a = dist(cov)
    nt.assert_allclose(a, np.array([[0.000000e+00, 6.552017e-01], [6.552017e-01, 0.0]]), rtol=1e-6, atol=1e-6)


def test_hrp():
    # use a small covariance matrix
    cov = np.array([[1, 0.5, 0], [0.5, 2, 0.0], [0, 0, 3]])

    # we compute the rootnode of a graph here
    # The rootnode points to left and right and has an id attribute.
    rootnode, link = root(dist(cov), 'single')

    v, w = hrp_feed(rootnode, cov=cov)
    nt.assert_allclose(v, np.linalg.multi_dot([w, cov, w]))
    nt.assert_allclose(w.sum(), 1.0)

    # risk parity in the branch...
    nt.assert_approx_equal(w[0], w[1] * 2)

    wi = np.array([2, 1]) / 3

    v1 = 0.8888888888888888
    v2 = 3.0
    nt.assert_approx_equal(1 * wi[0] ** 2 + 2 * wi[1] ** 2 + wi[0] * wi[1], v1)
    nt.assert_approx_equal(cov[2][2] * 1.0, v2)

    alpha, beta = split(v1, v2)
    nt.assert_approx_equal(alpha, 0.7714285714285715)
    nt.assert_approx_equal(v1 * alpha, v2 * beta)


def test_split():
    v1 = 3
    v2 = 5
    x1, x2 = split(v_left=v1, v_right=v2)
    nt.assert_allclose(np.array([x1, x2]), np.array([5.0, 3.0]) / 8.0)
    nt.assert_approx_equal(x1 + x2, 1.0)


def test_add():
    x = [1, 2, 3]
    y = [4, 5]
    assert x + y == [1, 2, 3, 4, 5]
